import traceback

from django.conf import settings
from django.contrib.gis.geos import GEOSGeometry

from cartodb import CartoDBAPIKey, CartoDBException

from .models import SyncEntry


class BaseSynchronizer(object):
    """
    The base class for synchronizing a model with a CartoDB table.

    The general outline for this process is:
    * Get model instances to synchronize
    * Synchronize using
    * Mark as not needing to be synchronized anymore

    INSERT and UPDATE statements inspired by:
     http://blog.cartodb.com/post/53301057653/faster-data-updates-with-cartodb
    """

    def __init__(self, model, cartodb_table, batch_size=500):
        self.model = model
        self.cartodb_table = cartodb_table
        self.batch_size = batch_size

    def synchronize(self):
        entries = self.get_entries_to_synchronize()
        try:
            self.synchronize_entries(entries)
            entries.mark_as_success()
        except Exception as e:
            traceback.print_exc()
            print('Exception while synchronizing:', e)
            # Assume none of them were updated
            entries.mark_as_failed()

    def get_entries_to_synchronize(self):
        entries = SyncEntry.objects.for_model(self.model).to_synchronize()
        entries = entries.order_by('-updated')
        if entries.count() > self.batch_size:
            entries = entries[:self.batch_size]
        return entries

    def synchronize_entries(self, entries):
        cl = CartoDBAPIKey(settings.CARTODB_SYNC['API_KEY'],
                            settings.CARTODB_SYNC['DOMAIN'])

        #
        # Apply synchronizations in the following order: insert, update,
        # delete. In this way, we should hopefully avoid unsafe situations
        # such as deleting an entry then trying to update it.
        #

        inserts = [self.get_cartodb_mapping(e.content_object) for e in entries \
                   if e.status == SyncEntry.PENDING_INSERT]
        if inserts:
            insert_statement = self.get_insert_statement(inserts)
            try:
                print(cl.sql(insert_statement))
            except CartoDBException as e:
                traceback.print_exc()
                print('Exception while inserting:', e)

        updates = [self.get_cartodb_mapping(e.content_object) for e in entries \
                   if e.status == SyncEntry.PENDING_UPDATE]
        if updates:
            statement = self.get_update_statement(updates)
            print(statement)
            try:
                print(cl.sql(statement))
            except CartoDBException as e:
                traceback.print_exc()
                print('Exception while updating:', e)

        deletes = [e for e in entries if e.status == SyncEntry.PENDING_DELETE]
        if deletes:
            statement = self.get_delete_statement(deletes)
            try:
                print(cl.sql(statement))
            except CartoDBException as e:
                traceback.print_exc()
                print('Exception while deleting:', e)

    def format_value(self, value):
        if value is None:
            return 'NULL'
        if isinstance(value, (int, float, complex)):
            return str(value)
        try:
            # Try converting to a number, avoid wrapping in quotes if it is one
            float(value)
            return str(value)
        except Exception:
            pass

        if isinstance(value, GEOSGeometry):
            return "'%s'" % value.ewkt
        return "'%s'" % str(value)

    def get_column_names(self):
        raise NotImplementedError('Implement get_column_names')

    def get_cartodb_mapping(self, instance):
        raise NotImplementedError('Implement get_cartodb_mapping')

    def get_delete_statement(self, delete_entries):
        """
        Consolidate the given instances into one large and efficient delete
        statement.

        DELETE FROM <table name> WHERE id in (<ids>)
        """
        sql = 'DELETE FROM %s WHERE id in (%s)' % (
            self.cartodb_table,
            ','.join([str(e.object_id) for e in delete_entries]),
        )
        return sql

    def get_insert_statement(self, insert_instances):
        """
        Consolidate the given instances into one large and efficient insert
        statement.

        INSERT INTO <table name> (<column names>) VALUES <values>
        """
        column_names = self.get_column_names()
        values = []
        for instance in insert_instances:
            formatted_values = [self.format_value(instance[c]) for c in column_names]
            values.append(','.join(formatted_values))

        sql = 'INSERT INTO %s (%s) VALUES %s' % (
            self.cartodb_table,
            ','.join(column_names),
            ','.join(['(%s)' % v for v in values]),
        )
        return sql

    def get_update_statement(self, instances):
        """
        Consolidate the given instances into one large and efficient update
        statement.

        UPDATE <table name> o
        SET <column 1>=n.<column 1>, <column 2>=n.<column 2>
        FROM (VALUES
            (<values in order for row 1>),
            (<values in order for row 2>),
            (<values in order for row 3>)
        ) n(id, <column 1>, <column 2>)
        WHERE o.id = n.id;
        """
        column_names = self.get_column_names()
        values = []
        for instance in instances:
            formatted_values = [self.format_value(instance[c]) for c in column_names]
            values.append(','.join(formatted_values))

        sql = 'UPDATE %s o SET %s FROM (VALUES %s) n(%s) WHERE o.id = n.id' % (
            self.cartodb_table,
            ','.join(['%s=n.%s' % (c, c) for c in column_names]),
            ','.join(['(%s)' % v for v in values]),
            ','.join(column_names),
        )
        return sql



from importlib import import_module
from optparse import make_option
import traceback

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.apps import apps

from ...models import SyncEntry


class Command(BaseCommand):
    help = 'Synchronize model data with CartoDB tables'

    option_list = BaseCommand.option_list + (
        make_option('--model',
            action='store',
            type='string',
            dest='model',
        ),
        make_option('--markinsert',
            action='store_true',
            dest='mark_insert',
        ),
    )

    def handle(self, *args, **options):
        model = options.get('model', None)
        mark_insert = options.get('mark_insert', False)

        try:
            models = settings.CARTODB_SYNC['MODELS']
            for m in models:
                if model and model != m['MODEL_CLASS']:
                    continue

                if mark_insert:
                    self.mark_all_as_pending_insert(m)
                    continue
                else:
                    self.synchronize(m)
        except Exception:
            traceback.print_exc()
            raise CommandError('There was a problem synchronizing data')

    def mark_all_as_pending_insert(self, settings_entry):
        model = self.get_model(settings_entry['MODEL_CLASS'])
        SyncEntry.objects.mark_as_pending_insert(model.objects.all())

    def get_model(self, model_string):
        return apps.get_model(*model_string.split('.'))

    def get_synchronizer_class(self, synchronizer_string):
        module_name, class_name = synchronizer_string.rsplit('.', 1)
        return getattr(import_module(module_name), class_name)

    def synchronize(self, settings_entry):
        model = self.get_model(settings_entry['MODEL_CLASS'])
        synchronizer_class = self.get_synchronizer_class(settings_entry['SYNCHRONIZER_CLASS'])
        synchronizer = synchronizer_class(model, settings_entry['CARTODB_TABLE'])
        synchronizer.synchronize()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='SyncEntry',
            fields=[
                ('id', models.AutoField(primary_key=True, verbose_name='ID', serialize=False, auto_created=True)),
                ('object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('status', models.CharField(max_length=20, blank=True, null=True, choices=[('pending delete', 'pending delete'), ('pending insert', 'pending insert'), ('pending update', 'pending update'), ('success', 'success'), ('fail', 'fail')])),
                ('added', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('attempts', models.IntegerField(default=0)),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType', blank=True, null=True)),
            ],
        ),
    ]



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.query import QuerySet


class SyncEntryQuerySet(QuerySet):

    def for_model(self, model):
        return self.filter(content_type=ContentType.objects.get_for_model(model))

    def to_synchronize(self):
        return self.filter(status__in=(SyncEntry.PENDING_DELETE,
                                       SyncEntry.PENDING_INSERT,
                                       SyncEntry.PENDING_UPDATE))

    def mark_as_status(self, instances, status):
        sync_entries = []
        for instance in instances:
            sync_entries.append(SyncEntry(content_object=instance, status=status))
        self.bulk_create(sync_entries)

    def mark_as_pending_delete(self, instances):
        # Delete other entries for these instances since they'll be deleted
        # soon anyway. Also potentially an error if the synchronizer will
        # attempt to access instances that no longer exist due to deletion.
        self.filter(object_id__in=[i.pk for i in instances]).delete()

        self.mark_as_status(instances, SyncEntry.PENDING_DELETE)

    def mark_as_pending_insert(self, instances):
        self.mark_as_status(instances, SyncEntry.PENDING_INSERT)

    def mark_as_pending_update(self, instances):
        # Delete other update entries for these instances
        pks = [instance.pk for instance in instances]
        self.filter(object_id__in=pks, status=SyncEntry.PENDING_UPDATE).delete()

        # Now add update entries for these instances
        self.mark_as_status(instances, SyncEntry.PENDING_UPDATE)

    def mark_as_success(self):
        # Need to fetch entries again rather than use self since it has likely
        # been sliced
        pks = [e.pk for e in self]
        SyncEntry.objects.filter(pk__in=pks).update(status=SyncEntry.SUCCESS)

    def mark_as_failed(self):
        for entry in self:
            entry.attempts = entry.attempts + 1
            if entry.attempts > 3:
                entry.status = SyncEntry.FAIL
            entry.save()


class SyncEntryManager(models.Manager):

    def get_queryset(self):
        return SyncEntryQuerySet(self.model)

    def __getattr__(self, name):
        return getattr(self.get_queryset(), name)


class SyncEntry(models.Model):
    # The object to be synchronized
    content_type = models.ForeignKey(ContentType, null=True, blank=True)
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    # Meta
    PENDING_DELETE = 'pending delete'
    PENDING_INSERT = 'pending insert'
    PENDING_UPDATE = 'pending update'
    SUCCESS = 'success'
    FAIL = 'fail'
    STATUS_CHOICES = (
        (PENDING_DELETE, 'pending delete'),
        (PENDING_INSERT, 'pending insert'),
        (PENDING_UPDATE, 'pending update'),
        (SUCCESS, 'success'),
        (FAIL, 'fail'),
    )
    status = models.CharField(max_length=20, choices=STATUS_CHOICES, null=True,
                              blank=True)
    added = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    attempts = models.IntegerField(default=0)

    objects = SyncEntryManager()



from setuptools import setup, find_packages
import os

import cartodbsync


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Web Environment',
    'Framework :: Django',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Software Development',
]

setup(
    author='Eric Brelsford',
    author_email='ebrelsford@gmail.com',
    name='django-cartodb-sync',
    version=cartodbsync.__version__,
    description=("A Django app that helps keep CartoDB tables synchronized with Django models."),
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),
    url='https://github.com/ebrelsford/django-cartodb-sync',
    license='BSD License',
    platforms=['OS Independent'],
    classifiers=CLASSIFIERS,
    install_requires=[
        'Django>=1.8',
        'cartodb>=0.8',
    ],
    packages=find_packages(),
    include_package_data=True,
)
